(* Mathematica Package *)
(* Created by the Wolfram Workbench Nov 11, 2010 *)

BeginPackage["DrosteEffect`"]

DrosteRadii::usage = "Radii is an option which specifies the two radii of the DrosteEffect function. \
The option-value must be in the form {r1,r2} where r1>0 and r2>r1. If an image is transformed with \
DrosteEffect then everything in the circular region with the radii r1 and r2 is used for the output\
 image."

DrosteEffect::usage = "DrosteEffect[img] calculates the so-called Droste-Effect of the image img. \
The method can be used with different implementations which can be selected through the \
DrosteImplementation option."

DrosteImplementation::usage = "DrosteImplementation is an option for DrosteEffect to select the \
used method. Currently three different values can be used: Automatic uses the straight \
forward Mathematica implementation. Compiled uses a compiled version of the same \
transform and WolframLinkLibrary uses a C implementation loaded through a library-call \
into Mathematica. "

DrosteCenter::usage = "TransformationCenter is an option for DrosteEffect which specifies \
the center of the transformation in image coordinates. Default is the center of the image."

DrosteRegion::usage = "TransformationRegion is an option for DrosteEffect which specifies \
the region of the transformation in image coordinates."

WolframLinkLibrary::usage = "WolframLinkLibrary is an option value for DrosteEffect which specifies \
to use the compiled shared library implementation "

DrosteEffect::invpar = "The optionvalue \"`1`\" of the option `2` is not of appropriate format."

Begin["`Private`"]

(* initialize the compiled library *)
$drosteLibrary = "drosteWL";

registerLibraryFunction[name_Symbol, cFunc_String, args_List, ret_] :=
   Set[name, LibraryFunctionLoad["drosteWL", cFunc, args, ret]];
registerLibraryFunction[funcs : {{_, _String, _List, _} ..}] := 
 Do[registerLibraryFunction @@ f,
  {f, funcs}]


$libfuncs = {{cLoadImage, 
    "loadImage", {{_Real, 2, "Shared"}}, _Integer}, {cUnloadImage, 
    "unloadImage", {}, _Integer}, {cInterpolate, 
    "interpolateImage", {_Real, _Real}, _Real}, {cDrosteImage, 
    "drosteImage", {_Real, _Real, _Real, _Real, _Real, _Real, _Real, 
        _Real, _Integer, _Integer}, {_Real, 2, Automatic}}};

registerLibraryFunction[$libfuncs];


Options[DrosteEffect] = {
  ImageSize -> Automatic,
  DrosteRadii -> Automatic,
  DrosteCenter -> Automatic,
  DrosteRegion -> Automatic,
  
  DrosteImplementation -> Automatic
  }
  

DrosteEffect[img_Image?(ImageChannels[#]===1&), OptionsPattern[]] := 
Module[{
  nx,ny,onx,ony, (* Image size and image output size *)
  r1,r2, (* inner and outer circle *)
  mx,my, (* center of the transformation *)
  bx1,bx2,by1,by2, (* bounding box the output image should occupy *)
  drosteFunc, (* the used implementation of the transform *)
  res
  },
   {nx,ny} = ImageDimensions[img];
   With[{outSize = OptionValue[ImageSize], def = {nx,ny}},
     {onx,ony} = Which[
       outSize === Automatic, def,
       Head[outSize]===Integer && outSize>1, {outSize,outSize},
       MatchQ[outSize,{x_Integer,y_Integer} /; x>1&&y>1], outSize,
       True, Message[DrosteEffect::invpar, outSize, ImageSize]; def
     ];  
   ];
   
   With[{radii=OptionValue[DrosteRadii], def = {0.1*Min[nx,ny],0.5*Min[nx,ny]}},
     {r1,r2} = Which[
       radii === Automatic, def,
       MatchQ[radii, {rr1_?NumericQ, rr2_?NumericQ}/; rr1>0.0&&rr2>rr1], radii,
       True, Message[DrosteEffect::invpar, radii, DrosteRadii]; def
     ];
   ];
   
   
   With[{cent=OptionValue[DrosteCenter], def = {nx/2,ny/2}},
     {mx,my} = Which[
       cent === Automatic, def,
       MatchQ[cent, {_?NumericQ, _?NumericQ}], cent,
       True, Message[DrosteEffect::invpar, cent, DrosteCenter]; def
     ];
   ];
   
   With[{bb=OptionValue[DrosteRegion], def = {{-nx,nx},{-ny,ny}}},
     {{bx1,bx2},{by1,by2}} = Which[
       bb === Automatic, def,
       MatchQ[bb, {{bbx1_?NumericQ, bbx2_?NumericQ},{bby1_?NumericQ,bby2_?NumericQ}}/;
         bbx1<bbx2&&bby1<bby2], bb,
       True, Message[DrosteEffect::invpar, bb, DrosteRegion]; def
     ];
   ]; 

   With[{impl=OptionValue[DrosteImplementation]},
     drosteFunc = Which[
       impl === Automatic, droste,
       impl === Compiled, drosteCompiled,
       impl === WolframLinkLibrary, drosteWolframLinkLibrary,
       True, Message[DrosteEffect::invpar, impl, DrosteImplementation]; droste
     ];
   ];
      
(*   Print[{r1,r2,-20nx,20nx,-20ny,20ny,mx,my,nx,ny,onx,ony}]; *)
   res=drosteFunc[ImageData[img,ImageType[img]],r1,r2,bx1,bx2,by1,by2,mx,my,nx,ny,onx,ony];
   If[MatrixQ[res],
     Image[res,ImageType[img]],
     $Failed
   ] 
];


DrosteEffect[img_Image, opts___?OptionQ] := ColorCombine[
    DrosteEffect[#,opts]&/@ColorSeparate[img]]

(* Implementation of the transformation which uses only normal Mathematica 
   functions without any optimization. The output image matrix is created sequentially
   without parallelization.
 *) 
droste[bm_?MatrixQ, r1_, r2_ , bx1_, bx2_, by1_, by2_, mx_, my_, nx_, ny_, outnx_, outny_] := 
Module[{x,y,ipimg,imgf},
    ipimg = ListInterpolation[bm, {{-my, ny-my},{-mx, nx-mx}}];
    imgf[z_Complex] := With[{xx = Re[z], yy = Im[z]},
        If[-mx <= xx <= nx-mx && -my <= yy <= ny-my, ipimg[yy, xx], 0.0]];
    Table[
      Block[{z = x + I y, r, phi, xx, yy, alpha, ff},
        (*
          if you don't know what happens in the next lines try google
          with the keywords "mathematics behind droste"
          Note that the transformation is backwards and the tiling is
          done in the Mod[...] step. 
        *)
        z = With[{out=Log[z]},If[!NumericQ[out],Log[z+$MinMachineNumber],out]]; (* catch Log[0.+0.0I]*)
          alpha = ArcTan[Log[r2/r1]/(2 Pi)];
          ff = Cos[alpha];
          z = z/ff*Exp[-I alpha];
          xx = Mod[Re[z], Log[r2/r1]];
          yy = Im[z];
          z = xx + I yy;
          imgf[r1*Exp[z]]
      ],{y,by1,by2,(by2-by1)/(outny-1)},{x,bx1,bx2,(bx2-bx1)/(outnx-1)}
    ]  
];

dfc = Compile[
   {{r1, _Real},
    {r2, _Real},
    {bx1, _Real},
    {bx2, _Real},
    {by1, _Real},
    {by2, _Real},
    {outnx, _Integer},
    {outny, _Integer}},
   Module[{x,y},
   Table[
    Block[{z = 0.0 + 0.0*I, r = 0.0, phi = 0.0, xx = 0.0, yy = 0.0, 
      alpha = 0.0, ff = 0.0, eps = 10.0^-6},
     If[x^2 + y^2 < eps,
      z = Log[$MinMachineNumber],
      z = Log[x + I y]];
     alpha = ArcTan[Log[r2/r1]/(2 Pi)];
     ff = Cos[alpha];
     z = z/ff*Exp[-I alpha];
     xx = Mod[Re[z], Log[r2/r1]];
     yy = Im[z];
     z = xx + I yy;
     r1*Exp[z]]
    ,
    {y, by1, by2, (by2 - by1)/(outny - 1)}, {x, bx1, 
     bx2, (bx2 - bx1)/(outnx - 1)}]   
   ]
];

(* Implementation of the transformation which uses only normal Mathematica 
   functions without any optimization. The output image matrix is created sequentially
   without parallelization.
 *) 
drosteCompiled[bm_?MatrixQ, r1_, r2_ , bx1_, bx2_, by1_, by2_, mx_, my_, nx_, ny_, outnx_, outny_] := 
Module[{ipimg,imgf},
    ipimg = ListInterpolation[bm, {{-my, ny-my},{-mx, nx-mx}}];
    imgf[z_Complex] := With[{xx = Re[z], yy = Im[z]},
        If[-mx <= xx <= nx-mx && -my <= yy <= ny-my, ipimg[yy, xx], 1.0]];
    Map[imgf,dfc[r1,r2,bx1,bx2,by1,by2,outnx,outny],{2}]
];

DrosteEffect::linkerr = "The Wolfram Link Library function `1` reported an error (`2`).";

drosteWolframLinkLibrary[bm_?MatrixQ, r1_, r2_ , bx1_, bx2_, by1_, by2_, 
	mx_, my_, (*nx*)_, (*ny*)_, outnx_, outny_]:=Module[{err, result},
	err = cLoadImage[bm];
	If[err =!= 0,
		Message[DrosteEffect::linkerr, cLoadImage, err];
		Return[$Failed]
	];
	(*r1_, r2_ , bx1_, bx2_, by1_, by2_, mx_, my_, outnx_, outny_*)
	result = cDrosteImage[r1,r2,bx1,bx2,by1,by2,mx,my,outnx,outny];
    err = cUnloadImage[];
    If[err =!= 0,
        Message[DrosteEffect::linkerr, cUnloadImage, err];
        Return[$Failed]
    ];
	result
]


End[]

EndPackage[]

