#include "WolframLibrary.h"
#include <cmath>

#define TINY (0.1e-12)

#ifndef M_PI
#define M_PI  3.14159265358979323846
#endif

struct mimage {
  WolframLibraryData libData;
  MTensor data;
  mint nx,ny;
  
  mimage(WolframLibraryData lib, MTensor imgData, mint nnx, mint nny):
    libData(lib),
    data(imgData),
    nx(nnx),
    ny(nny){}

  ~mimage() {
    data = 0;
    nx = 0;
    ny = 0;
  }

  mreal get(int x, int y) const {
    mint *pos = new mint[2]();
    pos[0] = y+1; pos[1] = x+1;
    double result = 0;
    libData->MTensor_getReal(data, pos, &result);
    delete pos;
    return result;
  }

  mreal getSave(int x, int y) const {
    if( x<0 || x>=nx || y<0 || y>=ny) return 0.0;
    return get(x,y);
  }

};

/**
 * Implementation of the Mitchell image interpolation with parameters
 * b=1/3 and c=1/3.
 */
class interpolator {
public:
  interpolator():support(2),b(0.333),c(0.333) {
    onesixth = 1.0/6.0;
    f11 = 12.0 - 9.0*b - 6*c;
    f12 = -18.0 + 12.0*b + 6.0*c;
    f13 = 6.0 - 2.0*b;
    
    f21 = -b - 6.0*c;
    f22 = 6.0*b + 30.0*c;
    f23 = -12.0*b - 48.0*c;
    f24 = 8.0*b + 24.0*c;
  }

  mreal valueAt(mreal x, mreal y, const mimage &img) {
    int xi = (int) floor(x),
      yi = (int) floor(y);
    mreal  xx = x - xi,
      yy = y - yi;
    mreal *wx = new mreal[2*support]();
  bool saveAccess = false;
  // Check whether we interpolate inside the image and can access the memory
  // without explicitly checking all the time.
  if(xi-support >=0 && xi+support < img.nx &&
     yi-support >=0 && yi+support < img.ny    )
    saveAccess = true;

  for(int i = 0; i<2*support; ++i) {
    wx[i] = getWeightAt(xx-support+i);
  }
  
  mreal (mimage::*getFunc)(mint, mint) const =
      saveAccess ? &mimage::get : &mimage::getSave;

  mreal res = 0.0;
  for(int j = 0; j<2*support; ++j) {
    mreal resx = 0.0;
    for(int i = 0; i<2*support; ++i)
      resx += (img.*getFunc)(xi+support-i, yi+support-j) * wx[i];
    res += resx*getWeightAt(yy-support+j);
  }
  delete [] wx;
  return res<0.0?0.0:res;
  }

private:
  int support;
  mreal a,b,c,f11,f12,f13,f21,f22,f23,f24,onesixth;
  mreal getWeightAt(mreal x) {
    x = fabs(x);
    mreal xx = x*x;
    if(x<1.0) return onesixth*(f11*xx*x + f12*xx + f13);
    if(x>=1 && x<2) return onesixth*(f21*xx*x + f22*xx + f23*x + f24);
    return 0.0;
  }
};

class droste {
public:
  static void transform(
        const mimage &source, // the source image
        mimage &target, // the target image
        mreal r1, // the inner radius in image pixel
        mreal r2, // the outer radius in image pixel
        mreal bx1, // bounding box
        mreal bx2,
        mreal by1,
        mreal by2,
        mreal xm, // x value of the transform center
        mreal ym // y value of the transform center
  );
private:
  droste();
  static inline mcomplex transformPoint(
      mcomplex pt, // the point to transform
      mreal r1, // the inner radius
      mreal r2, // the outer radius
      mreal alpha // the calculated alpha
      );
  static inline mcomplex transformPointID(mcomplex pt,mreal,mreal,mreal);
};


void droste::transform(
    const mimage &src,
    mimage &target,
    mreal r1,
    mreal r2,
    mreal bx1,
    mreal bx2,
    mreal by1,
    mreal by2,
    mreal mx,
    mreal my
    ){

  mint outnx = target.nx,
       outny = target.ny;
  mreal alpha = atan(log(r2/r1)/(2.0*M_PI));
  interpolator ip;
  mreal dx = (bx2-bx1)/(outnx-1);
  mreal dy = (by2-by1)/(outny-1);
  for(int j=0; j<outny; ++j)
    for(int i=0; i<outnx; ++i) {
      mcomplex pt;
      mcreal(pt) = bx1 + i*dx;
      mcimag(pt) = by1 + j*dy;
      pt = transformPoint(pt,r1,r2,alpha);
      mreal value = ip.valueAt(mcreal(pt)+mx, mcimag(pt)+my, src);
      mint pos[] = {j+1,i+1}; // Tensors start at 1!!
      target.libData->MTensor_setReal(target.data, pos, value);
    }
}

mcomplex droste::transformPointID(mcomplex pt, mreal, mreal, mreal){
  return pt;
}

mcomplex droste::transformPoint(mcomplex pt, mreal r1, mreal r2, mreal alpha) {
  mcomplex tmppt;

//   Log[z]
  mreal sqabsZ = (mcreal(pt)*mcreal(pt) + mcimag(pt)*mcimag(pt));
  if( sqabsZ < TINY ) {
    sqabsZ = 2.0*TINY*TINY;
    mcreal(pt) = TINY;
    mcimag(pt) = TINY;
  }
  mcreal(tmppt) = 0.5*log(sqabsZ);
  mcimag(tmppt) = atan2(mcimag(pt), mcreal(pt));
  mcreal(pt) = mcreal(tmppt);
  mcimag(pt) = mcimag(tmppt);

  // f*z*Exp[-I alpha]
  mreal cosalpha = cos(alpha);
  mreal sinalpha = sin(alpha);
  mreal f = 1.0/cosalpha;
  mcreal(tmppt) = f*mcreal(pt)*cosalpha + f*mcimag(pt)*sinalpha;
  mcimag(tmppt) = f*mcimag(pt)*cosalpha - f*mcreal(pt)*sinalpha;
  mcreal(pt) = mcreal(tmppt);
  mcimag(pt) = mcimag(tmppt);

  // tiling: re=Mod[Re[z], Log[r2/r1]]; im= Im[z]
  mreal logr2r1 = log(r2/r1);
  while(mcreal(pt) > logr2r1) mcreal(pt) -= logr2r1;
//
  // r1*Exp[z]
  mcreal(tmppt) = r1*exp(mcreal(pt))*cos(mcimag(pt));
  mcimag(tmppt) = r1*exp(mcreal(pt))*sin(mcimag(pt));
  return tmppt;
}

/**
 * Implementation of the library specific functions
 */


static mimage *globalImage = 0;
static interpolator *ip = 0;

DLLEXPORT mint WolframLibrary_getVersion() {
  return WolframLibraryVersion;
}

DLLEXPORT int WolframLibrary_initialize( WolframLibraryData libData) {
  return 0;
}

extern "C"
{

  DLLEXPORT int
  loadImage(WolframLibraryData libData, mint argc, MArgument *argv,
      MArgument res)
  {
    if (argc < 1) return LIBRARY_TYPE_ERROR;
    MTensor imgData = MArgument_getMTensor(argv[0]);
    mint type = libData->MTensor_getType(imgData);
    mint rank = libData->MTensor_getRank(imgData);
    mint const *dim = libData->MTensor_getDimensions(imgData);
    if (type != MType_Real) return LIBRARY_TYPE_ERROR;
    if (rank != 2) return LIBRARY_RANK_ERROR;
    if (dim[0] < 2 || dim[1] < 2) return LIBRARY_DIMENSION_ERROR;
    if(globalImage) delete globalImage;
    globalImage = new mimage(libData, imgData, dim[0], dim[1]);
    if(ip) delete ip;
    ip = new interpolator();
    MArgument_setInteger(res, 0);
    return LIBRARY_NO_ERROR;
  }

  DLLEXPORT int
  unloadImage(WolframLibraryData libData, mint argc, MArgument *argv,
      MArgument res)
  {
    if(globalImage) {
      libData->MTensor_disown(globalImage->data);
      delete globalImage;
      globalImage = 0;
    }
    if (ip) {
       delete ip;
       ip = 0;
    }
    MArgument_setInteger(res, 0);
    return LIBRARY_NO_ERROR;
  }

  DLLEXPORT int
  interpolateImage(WolframLibraryData libData, mint argc, MArgument *argv,
      MArgument res)
  {
    if (argc < 2) return LIBRARY_RANK_ERROR;
    mreal x = MArgument_getReal(argv[0]);
    mreal y = MArgument_getReal(argv[1]);
    MArgument_setReal(res, ip&&globalImage?ip->valueAt(x, y, *globalImage):-1.0);
    return LIBRARY_NO_ERROR;
  }

  // r1_, r2_ , bx1_, bx2_, by1_, by2_, mx_, my_, outnx_, outny_
  DLLEXPORT int
  drosteImage(WolframLibraryData libData, mint argc, MArgument *argv,
      MArgument res) {
    if(argc != 10) {
      libData->Message("Uncorrect number of parameter.");
      return LIBRARY_FUNCTION_ERROR;
    }
    if(globalImage == 0) {
      libData->Message("The source image was not initialized.");
      return LIBRARY_MEMORY_ERROR;
    }


    mreal r1 = MArgument_getReal(argv[0]);
    mreal r2 = MArgument_getReal(argv[1]);
    mreal bx1 = MArgument_getReal(argv[2]);
    mreal bx2 = MArgument_getReal(argv[3]);
    mreal by1 = MArgument_getReal(argv[4]);
    mreal by2 = MArgument_getReal(argv[5]);
    mreal mx = MArgument_getReal(argv[6]);
    mreal my = MArgument_getReal(argv[7]);
    mint outnx = MArgument_getInteger(argv[8]);
    mint outny = MArgument_getInteger(argv[9]);
    MTensor targetData;
    mint dim[] = {outny,outnx};
    mint err = libData->MTensor_new(MType_Real, 2, dim, &targetData);
    if(err != LIBRARY_NO_ERROR) {
      libData->Message("Could not create output image.");
      return LIBRARY_MEMORY_ERROR;
    }
    mimage *target = new mimage(libData, targetData, outnx, outny);
    droste::transform(*globalImage, *target, r1, r2, bx1, bx2, by1,
        by2, mx, my);
    MArgument_setMTensor(res,targetData);
    delete target;
    return LIBRARY_NO_ERROR;
  }

  DLLEXPORT int
  returnArray(WolframLibraryData libData, mint argc, MArgument *argv, MArgument res) {
    MTensor t;
    mint dim[] = {5,5};
    mint err = libData->MTensor_new(MType_Real, 2, dim, &t);
    if(err != LIBRARY_NO_ERROR) {
      libData->Message("Could not create tensor.");
      return LIBRARY_MEMORY_ERROR;
    }
    for(int j=0; j<5; ++j)
      for(int i=0; i<5; ++i) {
        mint pos[] = {i,j};
        libData->MTensor_setReal(t,pos,2.0 );
      }
    MArgument_setMTensor(res,t);
    return LIBRARY_NO_ERROR;
  }


} // extern "C"
